<?php
/**
 * @file
 * Provide Drush integration for release building and dependency building.
 */

/**
 * Implements hook_drush_command().
 */
function project_composer_drush_command() {
  $items = array();

  $items['project-composer-process-project'] = array(
    'description' => 'Build composer metadata for a given project shortname.',
    'callback' => 'drush_project_composer_process_project',
    'arguments' => array(
      'shortname' => 'Project shortname/uri',
    ),
    'aliases' => array('pcpp'),
  );
  $items['project-composer-process-namespaces'] = array(
      'description' => 'Build namespace mappings for project/module to packagename.',
      'callback' => 'drush_project_composer_process_namespaces',
      'aliases' => array('pcpn'),
  );
  $items['project-composer-manage-static-files'] = array(
    'description' => 'Cleans up static files queued for deletion',
    'callback' => 'drush_project_composer_manage_static_files',
    'aliases' => array('pcmf'),
  );
  return $items;
}


/**
 * Update and save all metadata for the project named.
 * accepts any number of project file names as arguments.
 */
function drush_project_composer_process_project() {
  $args = func_get_args();
  $provider_includes_to_rebuild = [];

  foreach ($args as $shortname) {
    drush_print("Processing package metadata for $shortname");
    // Get all versions for a project and build dependencies for them.
    $pid = project_get_nid_from_machinename($shortname);

    // Process the most recent release of a project for each major api.
    // Ideally this is future proofed (works with d9)
    foreach (taxonomy_get_tree(variable_get('project_release_api_vocabulary', '')) as $term) {
      if ($term->tid > 87) {
        // Allow processing for drupal/drupal for d7
        if (($term->tid > 103) && ($pid == 3060)) {
          continue;
        }
        $release_nodes = project_release_query_releases($pid, $term->tid);
        if ($release_nodes) {
          // Pop off the most recent release.
          $release_node = array_shift($release_nodes);
          $provider_includes_to_rebuild += project_composer_process_release($release_node);

          $wrapper = entity_metadata_wrapper('node', $release_node);
          $version = $wrapper->field_release_vcs_label->value();
          drush_print_r("Processed $shortname: {$version}");
        } else {
          drush_print("Skipped $shortname: {$term->name}");
        }
      }
    }
  }
  project_composer_manage_provider_includes($provider_includes_to_rebuild);
}

/**
 *
 */
function drush_project_composer_process_namespaces() {
  module_load_include('inc', 'project_usage', 'includes/date_api');

  drush_print("Processing package namespaces");

  $packages = project_composer_get_components();
  $existing_namespaces = project_composer_get_namespace_map();
  $existing_namespaces = [];
  // Pre-fill the namespace lookup table with drupal core modules.
  foreach (taxonomy_get_tree(variable_get('project_release_api_vocabulary', '')) as $term) {
    if ($term->tid > 87) {
      $core_module_names = project_composer_get_core_modules($term->tid);
      $core_version_string = project_composer_core_version_string_from_tid($term->tid);
      $drupal_packagename = project_composer_get_drupal_packagename($core_version_string);
      foreach ($core_module_names as $core_module) {
        $module['api_tid'] = $term->tid;
        $module['component_name'] = $core_module;
        $module['project_name'] = 'drupal';
        $module['package_namespace'] = substr($drupal_packagename, 7);
        $module['project_nid'] = '3060';
        $existing_namespaces[$term->tid][$core_module] = (object) $module;
        unset($module);
      }
    }
  }

  // Each record represents one project:component combination
  foreach ($packages as $package) {
    $core_version_string = project_composer_core_version_string_from_tid($package->api_tid);
    $package->component_name = drupal_strtolower($package->component_name);
    // Save the project as a namespace first
    $project_package = clone $package;
    $project_package->package_namespace = $package->project_name;
    $project_package->component_name = $package->project_name;

    if (isset($existing_namespaces[$project_package->api_tid][$project_package->project_name])) {
      // If we have already stored the project name as a package namespace, then either
      // 1. Package namespace is already taken by another project's submodule, or
      // 2. we've already processed at least one component of the current project

      $existing_package_project = $existing_namespaces[$project_package->api_tid][$project_package->project_name]->project_name;
      if (!($existing_package_project == $project_package->project_name)) {
        if ($existing_package_project != 'drupal') {
          // This could happen if we previously processed a component that happens
          // to match a project. For examaple, scald module contains a submodule
          // called scald_youtube.  There is also a scald_youtube module, but the
          // Namespace would already be taken.

          // Move the colliding component to project-component for its package namespace
          $new_namespace = $existing_namespaces[$project_package->api_tid][$project_package->project_name]->project_name . "-" . $existing_namespaces[$project_package->api_tid][$project_package->project_name]->package_namespace;

          $existing_namespaces[$project_package->api_tid][$new_namespace] = clone $existing_namespaces[$project_package->api_tid][$project_package->project_name];
          $existing_namespaces[$project_package->api_tid][$new_namespace]->package_namespace = $new_namespace;
          $existing_namespaces[$project_package->api_tid][$project_package->project_name] = $project_package;

          drush_print("D$core_version_string Project $project_package->project_name Collided with $new_namespace");

        } else {
          drush_print("D$core_version_string Core collision with Project $project_package->project_name");
          // If a project name collides with an existing core module, core trumps the project.
          // Initially, this should only be projects that were moved into core. Potentially it could
          // end up being modules that are moved out of core, but from the outset, we'll only start with a list
          // of core modules from the most recent version.

          // In other words, for d8, there isnt really a core 'entity' module namespace, because its not useful.
          $project_package->package_namespace = $project_package->project_name . "-" . $project_package->project_name;
        }
      }
      else {
        // We've processed this project earlier in the loop. noop.
      }
    }
    else {
      // If the package namespace doesnt yet exist for this project name, add it.
      $existing_namespaces[$project_package->api_tid][$project_package->project_name] = $project_package;
    }

    if (!isset($existing_namespaces[$package->api_tid][$package->component_name])) {
      // New record
      $package->package_namespace = $package->component_name;
      $existing_namespaces[$package->api_tid][$package->component_name] = $package;
    } else {
      // Record is either a duplicate, or it is a collision
      // If its a duplicate, skip it:
      $existing_package_project = $existing_namespaces[$package->api_tid][$package->component_name]->project_name;
      if (!($existing_package_project == $package->project_name)) {
        $package->package_namespace = $package->project_name . "-" . $package->component_name;
        drush_print("D$core_version_string Component: " . $package->project_name . ":" . $package->component_name . " Collided with " . $existing_package_project . ":" . $existing_namespaces[$package->api_tid][$package->component_name]->package_namespace . " Now: " . $package->package_namespace);
        if (!isset($existing_namespaces[$package->api_tid][$package->package_namespace])){
          $existing_namespaces[$package->api_tid][$package->package_namespace] = $package;
        }
      }
    }
  }
  foreach ($existing_namespaces as $apis){
    foreach ($apis as $package) {
      drupal_write_record('project_composer_namespace_map',$package);
    }
  }
}

/**
 *
 */
function drush_project_composer_manage_static_files() {
  $wait_files = [];

  drush_print("Processing static composer files");
  $queue = DrupalQueue::get('project_composer_file_expiration');
  drush_log(dt('@number files in queue.', array('@number' => $queue->numberOfItems())), 'ok');

  $timeout = drush_get_option('timeout', 0);
  $end = time() + $timeout;
  while ($item = $queue->claimItem()) {
    // Keep files for at least five minutes

    $item_age = time() - $item->data['timestamp'];
    if ($item_age > 1800) {
      drush_log(dt('Deleting file: @file', array('@file' => $item->data['filename'])), 'ok');
      file_unmanaged_delete($item->data['filename']);
      $filename = $item->data['filename'];
      $filename = preg_replace('|public:/|', 'files', $filename);
      $url = url($filename, array('absolute' => TRUE));
      drupalorg_fastly_purge_url($url);
      $queue->deleteItem($item);
    } else {
      drush_log(dt('File too fresh: @file @seconds', array('@file' => $item->data['filename'], '@seconds' => $item_age)), 'ok');
      $wait_files[] = $item;
    }
  }

  // Release any items we popped off the queue until the next cron run
  foreach ($wait_files as $item) {
    $queue->releaseItem($item);
  }

}

/**
 * Returns an array of every component in the project_database,
 * sorted by usage stat count in descending order.
 *
 * @return mixed
 */
function project_composer_get_components() {
  // Max timestamp subquery:
  $subquery = db_select('project_usage_week_project', 'puwp2');
  $subquery->addExpression('MAX(puwp2.timestamp)');

  $query = db_select('node', 'project_node')->extend('TableSort');
  $query->leftJoin('field_data_field_project_machine_name', 'fdfpmn', 'fdfpmn.entity_id = project_node.nid');
  $query->leftJoin('field_data_field_release_project', 'fdfrp', 'fdfrp.field_release_project_target_id = project_node.nid');
  $query->leftJoin('node', 'release_node', 'release_node.nid = fdfrp.entity_id');
  $query->leftJoin('project_dependency_component', 'pdc', 'pdc.release_nid = release_node.nid');
  $query->leftJoin('field_data_taxonomy_vocabulary_6', 'fdtv6', 'fdtv6.entity_id = release_node.nid');
  $query->leftJoin('project_usage_week_project', 'puwp', 'puwp.nid = project_node.nid AND puwp.timestamp = (' . $subquery . ') AND puwp.tid = fdtv6.taxonomy_vocabulary_6_tid');
  $query->addField('fdfpmn', 'field_project_machine_name_value', 'project_name');
  $query->addField('project_node', 'nid', 'project_nid');
  $query->addField('pdc', 'name', 'component_name');
  $query->addField('fdtv6', 'taxonomy_vocabulary_6_tid', 'api_tid');
  $query->addField('puwp', 'count', 'count');
  $query->condition('pdc.component_id', NULL, 'IS NOT NULL')
    ->condition('fdtv6.taxonomy_vocabulary_6_tid', 87, '>')
    // Hack for drupal camp colorado. Thanks fellas.
    ->condition('project_node.nid', 1797098, '!=')
    ->condition('project_node.type', ['project_module','project_theme'], 'IN')
    ->condition('project_node.status', 1, '=')
    ->distinct()
    ->orderBy('puwp.count', 'DESC')
    ->orderBy('fdtv6.taxonomy_vocabulary_6_tid');

  $packages = $query->execute()
    ->fetchAll();
  return $packages;
}

/**
 * Pull in all existing namespace mappings so that when we're looping over the components looking for new
 * namespaces to map, we do not in-advertantly change colliding components if one surpasses the other in
 * usage stats
 *
 * @return array
 */
function project_composer_get_namespace_map() {

  $existing_namespaces = [];
  $query = db_select('project_composer_namespace_map', 'pcnm');
  $namespace_map = $query->fields('pcnm')
    ->execute()
    ->fetchAll();
  foreach ($namespace_map as $record) {
    $existing_namespaces[$record->api_tid][$record->package_namespace] = $record;
  }
  return $existing_namespaces;
}
